#!/usr/bin/env python

"""
Inspect source files, obtaining details of classes and attributes.

Copyright (C) 2007, 2008, 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The results of inspecting a module are as follows:

Constants
---------

All constants identified within the code shall be registered.

Classes
-------

All global classes shall be registered; local classes (within functions) or
nested classes (within classes) are not currently registered.

Base classes must be detected and constant.

All classes without bases are made to inherit from __builtins__.object in order
to support some standard methods.

Functions
---------

All functions and lambda definitions shall be registered.

Namespaces
----------

Modules define their own "global" namespace, within which classes, functions
and lambda definitions establish a hierarchy of namespaces.

Only local, global and built-in namespaces are recognised; closures are not
supported.

Assignments
-----------

Name assignment and attribute assignment involving modules and classes cause
names to be associated with values within namespaces.

Any assignments within loops are considered to cause the targets of such
assignments to provide non-constant values.

Assignments to names are only really considered to cause the targets of such
assignments to provide constant values if the targets reside in the module
namespace or in class namespaces, subject to the above conditions.

Assignments to names within functions are not generally considered to cause the
targets of such assignments to provide constant values since functions can be
invoked many times with different inputs. This affects particularly the
definition of functions or lambdas within functions. However, there may be
benefits in considering a local to be constant within a single invocation.
"""

from micropython.common import *
from micropython.data import *
import compiler.ast
from compiler.visitor import ASTVisitor

# Program visitors.

class InspectedModule(ASTVisitor, Module):

    """
    An inspected module, providing core details via the Module superclass, but
    capable of being used as an AST visitor.
    """

    def __init__(self, name, importer):

        """
        Initialise this visitor with a module 'name' and an 'importer' which is
        used to provide access to other modules when required.
        """

        ASTVisitor.__init__(self)
        Module.__init__(self, name)
        self.visitor = self

        # Import machinery links.

        self.importer = importer
        self.optimisations = importer.optimisations
        self.builtins = self.importer.modules.get("__builtins__")
        self.loaded = 0

        # Current expression state.

        self.expr = None

        # Namespace state.

        self.in_init = 0        # Find instance attributes in __init__ methods.
        self.in_method = 0      # Find instance attributes in all methods.
        self.in_function = 0    # Note function presence, affecting definitions.
        self.in_loop = 0        # Note loop "membership", affecting assignments.
        self.namespaces = []
        self.module = None
        self.functions = []

    def parse(self, filename):

        "Parse the file having the given 'filename'."

        module = compiler.parseFile(filename)
        self.process(module)

    def process(self, module):

        "Process the given 'module'."

        self.astnode = self.module = module

        # Add __name__ to the namespace by adding an explicit assignment to the
        # module.

        module.node.nodes.insert(0, compiler.ast.Assign(
            [compiler.ast.AssName("__name__", 0)],
            compiler.ast.Const(self.full_name())
            ))

        # First, visit module-level code, recording global names.

        processed = self.dispatch(module)

        # Then, for each function, detect and record globals declared in those
        # functions. 

        for node, namespaces in self.functions:
            self.process_globals(node)

        self.finalise_namespace(self)

        # Then, visit each function, recording other names.

        for node, namespaces in self.functions:
            self._visitFunctionBody(node, namespaces)
            self.finalise_namespace(namespaces[-1])

        # Add references to other modules declared using the __all__ global.

        if self.has_key("__all__"):
            all = self["__all__"]
            if isinstance(all, compiler.ast.List):
                for n in all.nodes:
                    self.store(n.value, self.importer.add_module(self.name + "." + n.value))

        return processed

    def process_globals(self, node):

        """
        Within the given 'node', process global declarations, adjusting the
        module namespace.
        """

        for n in node.getChildNodes():
            if isinstance(n, compiler.ast.Global):
                for name in n.names:
                    if not self.has_key(name):
                        self[name] = None
            else:
                self.process_globals(n)

    def finalise_namespace(self, namespace):

        "Finalise the given 'namespace'."

        for names in namespace.get_all_attribute_usage():
            self.importer.use_names(names, namespace.full_name())

    def vacuum(self):

        """
        Vacuum the module namespace, removing unreferenced objects and unused
        names.
        """

        if self.should_optimise_unused_objects():
            self.vacuum_object(self)

            all_objects = list(self.all_objects)

            for obj in all_objects:
                if isinstance(obj, Class):
                    self.vacuum_object(obj)

    def vacuum_object(self, obj, delete_all=0):

        "Vacuum the given object 'obj'."

        for name, attr in obj.items():

            # Only consider deleting entire unused objects or things accessible
            # via names which are never used.

            if delete_all or not self.importer.uses_attribute(obj, name):
                del obj[name]

                # Delete any unambiguous attribute value. Such values can only
                # have been defined within the object and therefore are not
                # redefined by other code regions.

                if attr.assignments == 1:
                    value = attr.get_value()

                    if value is not obj and value in self.all_objects:
                        self.all_objects.remove(value)

                        # Delete class contents.

                        if isinstance(value, Class):
                            self.vacuum_object(value, 1)

    def unfinalise(self):

        "Reset finalised information for the module."

        for obj in self.all_objects:
            obj.unfinalise_attributes()

    def finalise(self):

        "Finalise the module."

        for obj in self.all_objects:
            obj.finalise_attributes()

    def add_object(self, obj, any_scope=0):

        """
        Record 'obj' if non-local or if the optional 'any_scope' is set to a
        true value.
        """

        if any_scope or not (self.namespaces and isinstance(self.namespaces[-1], Function)):
            self.all_objects.add(obj)

    # Optimisation tests.

    def should_optimise_unused_objects(self):
        return "unused_objects" in self.optimisations

    # Namespace methods.

    def in_class(self, namespaces=None):
        namespaces = namespaces or self.namespaces
        return len(namespaces) > 1 and isinstance(namespaces[-2], Class)

    def store(self, name, obj):

        "Record attribute or local 'name', storing 'obj'."

        if not self.namespaces:
            self.set(name, obj, not self.in_loop)
        else:
            self.namespaces[-1].set(name, obj, not self.in_loop)

    def store_lambda(self, obj):

        "Store a lambda function 'obj'."

        self.add_object(obj)

    def store_module_attr(self, name, module):

        """
        Record module attribute 'name' in the given 'module' using the current
        expression.
        """

        module.set(name, self.expr, 0)

    def store_class_attr(self, name):

        """
        Record class attribute 'name' in the current class using the current
        expression.
        """

        if self.in_method and self.namespaces[-2].has_key(name):
            self.namespaces[-2].set(name, self.expr, 0)
            return 1

        return 0

    def store_instance_attr(self, name):

        "Record instance attribute 'name' in the current class."

        if self.in_method:

            # Current namespace is the function.
            # Previous namespace is the class.

            self.namespaces[-2].add_instance_attribute(name)

    def get_namespace(self):

        "Return the parent (or most recent) namespace currently exposed."

        return (self.namespaces[-1:] or [self])[0]

    def use_name(self, name, node=None):

        """
        Use the given 'name' within the current namespace/unit, either in
        conjunction with a particular object (if 'node' is specified and not
        None) or unconditionally.
        """

        if node is not None and isinstance(node, compiler.ast.Name):
            self.use_attribute(node.name, name)

        # For general name usage, declare usage of the given name from this
        # particular unit.

        else:
            unit = self.get_namespace()
            self.importer.use_name(name, unit.full_name())

    # Attribute usage methods.
    # These are convenience methods which refer to the specific namespace's
    # implementation of these operations.

    def new_branchpoint(self):
        self.get_namespace()._new_branchpoint()

    def new_branch(self, loop_node=None):
        self.get_namespace()._new_branch(loop_node)

    def abandon_branch(self):
        self.get_namespace()._abandon_branch()

    def shelve_branch(self):
        self.get_namespace()._shelve_branch()

    def merge_branches(self):
        self.get_namespace()._merge_branches()

    def define_attribute_user(self, node):

        """
        Define 'node' as the user of attributes, indicating the point where the
        user is defined.
        """

        self.get_namespace()._define_attribute_user(node)

    def use_attribute(self, name, attrname):

        "Note usage on the attribute user 'name' of the attribute 'attrname'."

        return self.get_namespace()._use_attribute(name, attrname)

    def use_specific_attribute(self, objname, attrname):

        """
        Note attribute usage specifically on 'objname' - an object which is
        known at inspection time - or in the current unit if 'objname' is None,
        nominating a specific attribute 'attrname'.

        This bypasses attribute user mechanisms.
        """

        from_name = self.get_namespace().full_name()
        objname = objname or from_name
        self.importer.use_specific_name(objname, attrname, from_name)

    # Visitor methods.

    def default(self, node, *args):
        raise InspectError(self.full_name(), node, "Node class %r is not supported." % node.__class__)

    def NOP(self, node):
        for n in node.getChildNodes():
            self.dispatch(n)
        return None

    def NOP_ABANDON(self, node):
        self.NOP(node)
        self.abandon_branch()

    def OP(self, node):
        for n in node.getChildNodes():
            self.dispatch(n)
        return Instance()

    # Generic support for classes of operations.

    def _visitUnary(self, node):

        "Accounting method for the unary operator 'node'."

        method = unary_methods[node.__class__.__name__]
        self.use_name(method, node)
        return self.OP(node)

    def _visitBinary(self, node):

        "Accounting method for the binary operator 'node'."

        left_method, right_method = binary_methods[node.__class__.__name__]
        self.use_name(left_method, node)
        self.use_name(right_method, node)
        return self.OP(node)

    def _visitFunction(self, node, name):

        """
        Return a function object for the function defined by 'node' with the
        given 'name'. If a lambda expression is being visited, 'name' should be
        None.
        """

        # Define the function object.

        function = Function(
            name,
            self.get_namespace(),
            node.argnames,
            node.defaults,
            (node.flags & 4 != 0),
            (node.flags & 8 != 0),
            self.in_loop or self.in_function,
            self,
            node
            )

        self.add_object(function, any_scope=1)

        # Make a back reference from the node for code generation.

        node.unit = function

        # Process the defaults.

        for n in node.defaults:
            self.expr = self.dispatch(n)
            function.store_default(self.expr)

        self.functions.append((node, self.namespaces + [function]))

        # Store the function.

        if name is not None:
            self.store(name, function)
        else:
            self.store_lambda(function)

        # Test the defaults and assess whether an dynamic object will result.

        function.make_dynamic()
        return function

    def _visitFunctionBody(self, node, namespaces):

        "Enter the function."

        # Current namespace is the function.
        # Previous namespace is the class.

        if self.in_class(namespaces):
            if namespaces[-1].name == "__init__":
                self.in_init = 1
            self.in_method = 1

        in_function = self.in_function
        in_loop = self.in_loop
        self.in_function = 1
        self.in_loop = 0

        self.namespaces = namespaces
        self.dispatch(node.code)

        self.in_loop = in_loop
        self.in_function = in_function
        self.in_init = 0
        self.in_method = 0

    # Specific handler methods.

    visitAdd = _visitBinary

    visitAnd = OP

    visitAssert = NOP

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssAttr(self, node):
        expr = self.dispatch(node.expr)

        # Record the attribute on the presumed target.

        if isinstance(expr, Attr):
            if expr.name == "self":
                if not self.store_class_attr(node.attrname):
                    self.store_instance_attr(node.attrname)
            elif isinstance(expr.get_value(), Module):
                self.store_module_attr(node.attrname, expr.get_value())
                print "Warning: attribute %r of module %r set outside the module." % (node.attrname, expr.get_value().name)

            # Note usage of the attribute where a local is involved.

            if expr.parent is self.get_namespace():
                node._attrusers = self.use_attribute(expr.name, node.attrname)
                node._username = expr.name

        return None

    def visitAssList(self, node):

        # Declare names which will be used by generated code.

        self.use_name("__getitem__", node)

        # Process the assignment.

        for i, n in enumerate(node.nodes):
            self.dispatch(n)
            self.importer.make_constant(i) # for __getitem__(i) at run-time
        return None

    def visitAssName(self, node):
        if hasattr(node, "flags") and node.flags == "OP_DELETE":
            print "Warning: deletion of attribute %r in %r is not supported." % (node.name, self.full_name())
            #raise InspectError(self.full_name(), node, "Deletion of attribute %r is not supported." % node.name)

        self.store(node.name, self.expr)
        self.define_attribute_user(node)

        # Ensure the presence of the given name in this namespace.

        self.use_specific_attribute(None, node.name)
        return None

    visitAssTuple = visitAssList

    def visitAugAssign(self, node):

        # Accounting.

        aug_method, (left_method, right_method) = augassign_methods[node.op]
        self.use_name(aug_method, node)
        self.use_name(left_method, node)
        self.use_name(right_method, node)

        # Process the assignment.

        self.expr = self.dispatch(node.expr)

        # NOTE: Similar to micropython.ast handler code.
        # NOTE: Slices and subscripts not supported.

        if isinstance(node.node, compiler.ast.Name):
            self.visitAssName(node.node)
        elif isinstance(node.node, compiler.ast.Getattr):
            self.visitAssAttr(node.node)
        else:
            raise InspectError(self.full_name(), node, "AugAssign(Slice or Subscript)")

        return None

    visitBackquote = OP

    visitBitand = _visitBinary

    visitBitor = _visitBinary

    visitBitxor = _visitBinary

    visitBreak = NOP_ABANDON

    visitCallFunc = OP

    def visitClass(self, node):

        """
        Register the class at the given 'node' subject to the restrictions
        mentioned in the module docstring.
        """

        if self.namespaces:
            print "Class %r in %r is not global: ignored." % (node.name, self.namespaces[-1].full_name())
            return None
        else:
            if self.in_loop:
                print "Warning: class %r in %r defined in a loop." % (node.name, self.full_name())

            cls = Class(node.name, self.get_namespace(), self, node)

            # Visit the base class expressions, attempting to find concrete
            # definitions of classes.

            for base in node.bases:
                expr = self.dispatch(base)
                if isinstance(expr, Attr):
                    if expr.assignments != 1:
                        raise InspectError(self.full_name(), node,
                            "Base class %r for %r is not constant." % (base, cls.full_name()))
                    else:
                        cls.add_base(expr.get_value())
                else: # if expr is None:
                    raise InspectError(self.full_name(), node,
                        "Base class %r for %r is not found: it may be hidden in some way." % (base, cls.full_name()))

            # NOTE: Potentially dubious measure to permit __init__ availability.
            # If no bases exist, adopt the 'object' class.

            if not node.bases and not (self.name == "__builtins__" and node.name == "object") :
                expr = self.dispatch(compiler.ast.Name("object"))
                cls.add_base(expr.get_value())

            # Make a back reference from the node for code generation.

            node.unit = cls

            # Make an entry for the class.

            self.store(node.name, cls)
            self.add_object(cls)

            # Process the class body.

            self.namespaces.append(cls)
            self.dispatch(node.code)
            self.finalise_namespace(cls)
            self.namespaces.pop()

            return cls

    def visitCompare(self, node):

        # Accounting.
        # NOTE: Replicates some code in micropython.ast.visitCompare.

        this_node = node

        for op in node.ops:
            op_name, next_node = op

            # Get the applicable methods.

            methods = comparison_methods[op_name]

            # Define name/attribute usage.

            if methods is not None:
                self.use_name(methods[0], this_node)
                self.use_name(methods[1], next_node)
            elif op_name.endswith("in"):
                self.use_name("__contains__", next_node)

            this_node = next_node

        return self.OP(node)

    def visitConst(self, node):

        # Register the constant, if necessary, returning the resulting object.

        self.use_name(self.importer.get_constant_type_name(node.value), node)
        return self.importer.make_constant(node.value)

    visitContinue = NOP_ABANDON

    visitDecorators = NOP

    visitDict = OP

    visitDiscard = NOP

    visitDiv = _visitBinary

    visitEllipsis = NOP

    visitExec = NOP

    visitExpression = OP

    visitFloorDiv = _visitBinary

    def visitFor(self, node):
        self.new_branchpoint()

        # Declare names which will be used by generated code.

        self.use_name("__iter__", node.list)
        self.use_name("next")

        in_loop = self.in_loop
        self.in_loop = 1
        self.dispatch(node.assign)
        self.dispatch(node.list)

        # Enter the loop.
        # Propagate attribute usage to branches.

        self.new_branch()
        self.dispatch(node.body)
        self.shelve_branch()
        self.in_loop = in_loop

        # Maintain a branch for the else clause or the current retained usage
        # where execution avoids the conditional clauses.

        self.new_branch()
        if node.else_ is not None:
            self.dispatch(node.else_)
        self.shelve_branch()

        self.merge_branches()
        return None

    def visitFrom(self, node):
        module = self.importer.load(node.modname, 1)

        #if module is None:
        #    print "Warning:", node.modname, "not imported."

        for name, alias in node.names:
            if name != "*":
                if module is not None and module.has_key(name):
                    attr = module[name]
                    self.store(alias or name, attr)
                    if isinstance(attr.get_value(), Module) and not attr.get_value().loaded:
                        self.importer.load(attr.get_value().name)

                # Support the import of names from missing modules.

                else:
                    self.store(alias or name, UnresolvedName(name, node.modname, self))
            else:
                if module is not None:
                    for n in module.keys():
                        attr = module[n]
                        self.store(n, attr)
                        if isinstance(attr.get_value(), Module) and not attr.get_value().loaded:
                            self.importer.load(attr.get_value().name)

        return None

    def visitFunction(self, node):
        return self._visitFunction(node, node.name)

    visitGenExpr = OP

    visitGenExprFor = NOP

    visitGenExprIf = NOP

    visitGenExprInner = NOP

    def visitGetattr(self, node):
        expr = self.dispatch(node.expr)
        attrname = node.attrname

        # Attempt to identify the nature of the attribute.

        if isinstance(expr, Attr):
            value = expr.get_value()

            # Get the attribute and record its usage.

            if isinstance(value, (Class, Module)):
                attr = value.get(attrname)
                self.use_specific_attribute(value.full_name(), attrname)

            elif isinstance(value, UnresolvedName):
                attr = UnresolvedName(attrname, value.full_name(), self)

            else:
                attr = None

                # Note usage of the attribute where a local is involved.

                if expr.parent is self.get_namespace():
                    node._attrusers = self.use_attribute(expr.name, attrname)
                    node._username = expr.name
                else:
                    self.use_name(attrname, node.expr)

        elif self.builtins is not None:
            attr = self.builtins.get(attrname)
            self.use_specific_attribute(self.builtins.full_name(), attrname)

        else:
            attr = UnresolvedName(attrname, value.full_name(), self)

        return attr

    def visitGlobal(self, node):
        if self.namespaces:
            for name in node.names:
                ns = self.namespaces[-1]
                if not ns.make_global(name):
                    raise InspectError(ns.full_name(), node, "Name %r is global and local in %r" % (name, ns.full_name()))

                # The name is recorded in an earlier process.

    def visitIf(self, node):
        self.new_branchpoint()

        # Propagate attribute usage to branches.

        for test, body in node.tests:
            self.dispatch(test)

            self.new_branch()
            self.dispatch(body)
            self.shelve_branch()

        # Maintain a branch for the else clause or the current retained usage
        # where execution avoids the conditional clauses.

        self.new_branch()
        if node.else_ is not None:
            self.dispatch(node.else_)
        self.shelve_branch()

        self.merge_branches()
        return None

    visitIfExp = NOP

    def visitImport(self, node):
        for name, alias in node.names:
            if alias is not None:
                module = self.importer.load(name, 1) or UnresolvedName(None, name, self)
                self.store(alias, module)
            else:
                module = self.importer.load(name) or UnresolvedName(None, name.split(".")[0], self)
                self.store(name.split(".")[0], module)

        return None

    visitInvert = _visitUnary

    def visitKeyword(self, node):
        self.dispatch(node.expr)
        self.importer.make_constant(node.name)
        self.keyword_names.add(node.name)
        return None

    def visitLambda(self, node):
        self.use_name(None) # lambda functions have no names but are assumed to be invoked
        return self._visitFunction(node, None)

    visitLeftShift = _visitBinary

    visitList = OP

    visitListComp = OP

    visitListCompFor = NOP

    visitListCompIf = NOP

    visitMod = _visitBinary

    def visitModule(self, node):

        # Make a back reference from the node for code generation.

        node.unit = self
        return self.dispatch(node.node)

    visitMul = _visitBinary

    def visitName(self, node):
        name = node.name

        # Constants.

        if self.importer.predefined_constants.has_key(name):
            attr = self.importer.get_predefined_constant(name)

        # Locals.

        elif self.namespaces and self.namespaces[-1].has_key(name):
            attr = self.namespaces[-1][name]

            # Note usage of the local (potentially a class attribute).

            self.use_specific_attribute(None, name)

        # Globals.

        elif self.has_key(name):
            attr = self[name]

            # Note usage of the module attribute.

            self.use_specific_attribute(self.full_name(), name)

        # Builtins.

        elif self.builtins is not None and self.builtins.has_key(name):
            attr = self.builtins[name]
            self.use_specific_attribute(self.builtins.full_name(), name)

        # Unknown.

        else:
            attr = None
            self.use_name(name)

        return attr

    visitNot = OP

    visitOr = OP

    visitPass = NOP

    visitPower = _visitBinary

    visitPrint = NOP

    visitPrintnl = NOP

    visitRaise = NOP_ABANDON

    visitReturn = NOP_ABANDON

    visitRightShift = _visitBinary

    visitSlice = OP

    visitSliceobj = OP

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    visitSub = _visitBinary

    def visitSubscript(self, node):
        self.use_name("__getitem__", node)
        self.OP(node)

    def visitTryExcept(self, node):
        self.dispatch(node.body)

        self.new_branchpoint()

        for name, var, n in node.handlers:
            self.new_branch()

            # Establish the local for the handler.

            if var is not None:
                self.dispatch(var)
            if n is not None:
                self.dispatch(n)

            self.shelve_branch()

        if node.else_ is not None:
            self.new_branch()
            self.dispatch(node.else_)
            self.shelve_branch()

        self.merge_branches()
        return None

    visitTryFinally = NOP

    visitTuple = OP

    visitUnaryAdd = _visitUnary

    visitUnarySub = _visitUnary

    def visitWhile(self, node):
        self.new_branchpoint()

        # Propagate attribute usage to branches.

        in_loop = self.in_loop
        self.in_loop = 1
        self.dispatch(node.test)
        self.new_branch(node)
        self.dispatch(node.body)
        self.shelve_branch()
        self.in_loop = in_loop

        # Maintain a branch for the else clause or the current retained usage
        # where execution avoids the conditional clauses.
        # NOTE: Consider merging here before the else clause.

        self.new_branch()
        if node.else_ is not None:
            self.dispatch(node.else_)
        self.shelve_branch()

        self.merge_branches()
        return None

    visitWith = NOP

    visitYield = NOP

# vim: tabstop=4 expandtab shiftwidth=4
